import io
from datetime import datetime
from pathlib import Path
from typing import Annotated, Optional

import typer
from dateutil.tz import tzlocal
from monty.json import jsanitize
from monty.serialization import dumpfn
from qtoolkit.core.data_objects import QResources
from rich.pretty import pprint

from jobflow_remote import SETTINGS
from jobflow_remote.cli.formatting import (
    format_job_info,
    get_job_info_table,
    get_job_report_components,
    header_name_data_getter_map,
)
from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.types import (
    OptionalStr,
    break_lock_opt,
    days_opt,
    db_ids_opt,
    delete_all_opt,
    delete_files_opt,
    delete_output_opt,
    end_date_opt,
    flow_ids_opt,
    hours_opt,
    job_db_id_arg,
    job_ids_indexes_opt,
    job_index_arg,
    job_index_opt,
    job_state_arg,
    job_state_opt,
    locked_opt,
    max_results_opt,
    metadata_opt,
    name_opt,
    query_opt,
    raise_on_error_opt,
    reverse_sort_flag_opt,
    sort_opt,
    start_date_opt,
    verbosity_opt,
    wait_lock_opt,
    worker_name_opt,
)
from jobflow_remote.cli.utils import (
    ReportInterval,
    SortOption,
    check_incompatible_opt,
    check_query_incompatibility,
    check_stopped_runner,
    execute_multi_jobs_cmd,
    exit_with_error_msg,
    exit_with_warning_msg,
    get_config_manager,
    get_job_controller,
    get_job_db_ids,
    get_job_ids_indexes,
    get_start_date,
    hide_progress,
    loading_spinner,
    out_console,
    print_success_msg,
    str_to_dict,
)
from jobflow_remote.jobs.report import JobsReport
from jobflow_remote.jobs.state import JobState
from jobflow_remote.remote.queue import ERR_FNAME, OUT_FNAME

app_job = JFRTyper(
    name="job", help="Commands for managing the jobs", no_args_is_help=True
)
app.add_typer(app_job)


@app_job.command(name="list")
def jobs_list(
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    worker_name: worker_name_opt = None,
    verbosity: verbosity_opt = 0,
    max_results: max_results_opt = 100,
    sort: sort_opt = SortOption.UPDATED_ON,
    reverse_sort: reverse_sort_flag_opt = False,
    locked: locked_opt = False,
    custom_query: query_opt = None,
    error: Annotated[
        bool,
        typer.Option(
            "--error",
            "-e",
            help="Select the jobs in FAILED and REMOTE_ERROR state. Incompatible with the --state option",
        ),
    ] = False,
    stored_data_keys: Annotated[
        Optional[list[str]],
        typer.Option(
            "--stored-data-key",
            "-sdk",
            help="Key to be shown from the stored_data field.",
        ),
    ] = None,
    cli_output_keys: Annotated[
        Optional[str],
        typer.Option(
            "--output",
            "-o",
            help=f"Table columns to be shown. Needs to be specified as string with comma separated keys, e.g."
            f"'state,db_id,name'. Overrides the verbosity option. Can also be set in the config file. "
            f"Available options are: {', '.join(header_name_data_getter_map)}",
        ),
    ] = None,
):
    """
    Get the list of Jobs in the database.
    """
    check_incompatible_opt({"start_date": start_date, "days": days, "hours": hours})
    check_incompatible_opt({"end_date": end_date, "days": days, "hours": hours})
    check_incompatible_opt({"state": state, "error": error})
    check_incompatible_opt({"output": cli_output_keys, "verbosity": verbosity})
    check_query_incompatibility(
        custom_query,
        [
            job_id,
            db_id,
            flow_id,
            state,
            start_date,
            end_date,
            name,
            metadata,
            days,
            hours,
            worker_name,
        ],
    )
    output_keys = (
        cli_output_keys.split(",")
        if cli_output_keys
        else SETTINGS.cli_job_list_columns or []
    )
    if not set(output_keys).issubset(header_name_data_getter_map):
        exit_with_error_msg(
            f"Header keys not supported: {set(output_keys).difference(header_name_data_getter_map)}"
        )
    if stored_data_keys and not set(output_keys).isdisjoint(stored_data_keys):
        exit_with_error_msg(
            "Specifying a stored data key which is a standard column is disallowed."
        )

    job_ids_indexes = get_job_ids_indexes(job_id)

    jc = get_job_controller()

    start_date = get_start_date(start_date, days, hours)

    db_sort: list[tuple[str, int]] = [(sort.value, 1 if reverse_sort else -1)]

    if error:
        state = [JobState.REMOTE_ERROR, JobState.FAILED]

    with loading_spinner():
        if custom_query:
            jobs_info = jc.get_jobs_info_query(
                query=custom_query,
                limit=max_results,
                sort=db_sort,
            )
        else:
            jobs_info = jc.get_jobs_info(
                job_ids=job_ids_indexes,
                db_ids=db_id,
                flow_ids=flow_id,
                states=state,
                start_date=start_date,
                locked=locked,
                end_date=end_date,
                name=name,
                metadata=metadata,
                workers=worker_name,
                limit=max_results,
                sort=db_sort,
            )

        table = get_job_info_table(
            jobs_info,
            verbosity=verbosity,
            output_keys=output_keys,
            stored_data_keys=stored_data_keys,
        )

    out_console.print(table)
    if SETTINGS.cli_suggestions:
        if max_results and len(jobs_info) == max_results:
            out_console.print(
                f"The number of Jobs printed may be limited by the maximum selected: {max_results}",
                style="yellow",
            )
        remote_errors = False
        if any(
            ji.remote.retry_time_limit is not None and ji.state != JobState.REMOTE_ERROR
            for ji in jobs_info
        ):
            text = (
                "Some jobs (state in orange) have failed while interacting with"
                " the worker, but will be retried."
            )
            out_console.print(text, style="yellow")
            remote_errors = True
        if remote_errors or any(
            ji.state in (JobState.REMOTE_ERROR, JobState.FAILED) for ji in jobs_info
        ):
            text = "Get more information about the errors with 'jf job info JOB_ID'"
            out_console.print(text, style="yellow")


@app_job.command(name="info")
def job_info(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    pid: Annotated[
        str,
        typer.Option(
            "--pid",
            help="The process ID of the job in the queue system (e.g. Slurm job ID)"
            "The -v flag will be ignored if --pid is used.",
        ),
    ] = None,
    show_none: Annotated[
        bool,
        typer.Option(
            "--show-none",
            "-n",
            help="Show the data whose values are None. Usually hidden",
        ),
    ] = False,
    verbosity: verbosity_opt = 0,
) -> None:
    """Detailed information on a specific job."""
    if pid is not None and job_db_id is not None:
        raise typer.BadParameter("Cannot specify both job ID/index and process ID")

    with loading_spinner():
        jc = get_job_controller()

        if job_db_id is not None:
            db_id, job_id = get_job_db_ids(job_db_id, job_index)

            if verbosity > 0:
                job_data = jc.get_job_doc(
                    job_id=job_id,
                    job_index=job_index,
                    db_id=db_id,
                )
            else:
                job_data = jc.get_job_info(
                    job_id=job_id,
                    job_index=job_index,
                    db_id=db_id,
                )
        elif pid is not None:
            job_data = jc.get_job_info_by_pid(pid)
        else:
            raise typer.BadParameter("Must specify either job ID/index or process ID")

    if not job_data:
        exit_with_error_msg("No data matching the request")

    out_console.print(format_job_info(job_data, verbosity, show_none=show_none))


@app_job.command()
def set_state(
    state: job_state_arg,
    job_db_id: job_db_id_arg,
    job_index: job_index_arg = None,
) -> None:
    """
    Sets the state of a Job to an arbitrary value.
    WARNING: No checks. This can lead to inconsistencies in the DB. Use with care.
    """
    db_id, job_id = get_job_db_ids(job_db_id, job_index)

    with loading_spinner():
        jc = get_job_controller()

        succeeded = jc.set_job_state(
            state=state,
            job_id=job_id,
            job_index=job_index,
            db_id=db_id,
        )

    if not succeeded:
        exit_with_error_msg("Could not change the job state")

    print_success_msg()


@app_job.command()
def rerun(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    worker_name: worker_name_opt = None,
    custom_query: query_opt = None,
    verbosity: verbosity_opt = 0,
    wait: wait_lock_opt = None,
    break_lock: break_lock_opt = False,
    force: Annotated[
        bool,
        typer.Option(
            "--force",
            "-f",
            help=(
                "Force the rerun even if some conditions would normally prevent it (e.g. "
                "state usually not allowed or children already executed). Can lead to "
                "inconsistencies. Advanced users."
            ),
        ),
    ] = False,
    raise_on_error: raise_on_error_opt = False,
    no_delete: Annotated[
        bool,
        typer.Option(
            "--no-delete",
            "-nd",
            help=("Skip the delete of the files on the worker."),
        ),
    ] = False,
) -> None:
    """
    Rerun a Job. By default, this is limited to jobs that failed and children did
    not start or jobs that are running. The rerun Job is set to READY and children
    Jobs to WAITING. If possible, the associated job submitted to the remote queue
    will be cancelled. Most of the limitations can be overridden by the 'force'
    option. This could lead to inconsistencies in the overall state of the Jobs of
    the Flow.
    All the folders of the Jobs whose state are modified will also be deleted on
    the worker.
    """
    if force or break_lock:
        check_stopped_runner(error=False)

    jc = get_job_controller()

    execute_multi_jobs_cmd(
        single_cmd=jc.rerun_job,
        multi_cmd=jc.rerun_jobs,
        job_db_id=job_db_id,
        job_index=job_index,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        states=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        workers=worker_name,
        custom_query=custom_query,
        days=days,
        hours=hours,
        verbosity=verbosity,
        wait=wait,
        break_lock=break_lock,
        force=force,
        raise_on_error=raise_on_error,
        interactive=True,
        delete_files=not no_delete,
    )


@app_job.command()
def retry(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    worker_name: worker_name_opt = None,
    custom_query: query_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    wait: wait_lock_opt = None,
    break_lock: break_lock_opt = False,
    raise_on_error: raise_on_error_opt = False,
) -> None:
    """
    Retry to perform the operation that failed for a job in a REMOTE_ERROR state
    or reset the number of attempts at remote action, in order to allow the
    runner to try it again immediately.
    """
    if break_lock:
        check_stopped_runner(error=False)

    jc = get_job_controller()

    execute_multi_jobs_cmd(
        single_cmd=jc.retry_job,
        multi_cmd=jc.retry_jobs,
        job_db_id=job_db_id,
        job_index=job_index,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        states=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        workers=worker_name,
        custom_query=custom_query,
        verbosity=verbosity,
        wait=wait,
        break_lock=break_lock,
        raise_on_error=raise_on_error,
    )


@app_job.command()
def pause(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    worker_name: worker_name_opt = None,
    custom_query: query_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    wait: wait_lock_opt = None,
    raise_on_error: raise_on_error_opt = False,
) -> None:
    """Pause a Job. Only READY and WAITING Jobs can be paused. The operation is reversible."""
    jc = get_job_controller()

    execute_multi_jobs_cmd(
        single_cmd=jc.pause_job,
        multi_cmd=jc.pause_jobs,
        job_db_id=job_db_id,
        job_index=job_index,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        states=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        workers=worker_name,
        custom_query=custom_query,
        verbosity=verbosity,
        wait=wait,
        raise_on_error=raise_on_error,
    )


@app_job.command()
def play(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    worker_name: worker_name_opt = None,
    custom_query: query_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    wait: wait_lock_opt = None,
    raise_on_error: raise_on_error_opt = False,
) -> None:
    """Resume a Job that was previously PAUSED."""
    jc = get_job_controller()

    execute_multi_jobs_cmd(
        single_cmd=jc.play_job,
        multi_cmd=jc.play_jobs,
        job_db_id=job_db_id,
        job_index=job_index,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        states=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        workers=worker_name,
        custom_query=custom_query,
        verbosity=verbosity,
        wait=wait,
        raise_on_error=raise_on_error,
    )


@app_job.command()
def stop(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    worker_name: worker_name_opt = None,
    custom_query: query_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    wait: wait_lock_opt = None,
    break_lock: break_lock_opt = False,
    raise_on_error: raise_on_error_opt = False,
) -> None:
    """
    Stop a Job. Only Jobs that did not complete or had an error can be stopped.
    The operation is irreversible.
    If possible, the associated job submitted to the remote queue will be cancelled.
    """
    if break_lock:
        check_stopped_runner(error=False)

    jc = get_job_controller()

    execute_multi_jobs_cmd(
        single_cmd=jc.stop_job,
        multi_cmd=jc.stop_jobs,
        job_db_id=job_db_id,
        job_index=job_index,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        states=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        workers=worker_name,
        custom_query=custom_query,
        verbosity=verbosity,
        wait=wait,
        break_lock=break_lock,
        raise_on_error=raise_on_error,
    )


@app_job.command()
def delete(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    worker_name: worker_name_opt = None,
    custom_query: query_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    wait: wait_lock_opt = None,
    raise_on_error: raise_on_error_opt = False,
    delete_output: delete_output_opt = False,
    delete_files: delete_files_opt = False,
    delete_all: delete_all_opt = False,
) -> None:
    """
    Delete Jobs individually. The Flow document will be updated accordingly but
    no consistency check is performed. The Flow may be left in an inconsistent state.
    For advanced users only.
    """

    if delete_all:
        delete_files = delete_output = True

    jc = get_job_controller()

    execute_multi_jobs_cmd(
        single_cmd=jc.delete_job,
        multi_cmd=jc.delete_jobs,
        job_db_id=job_db_id,
        job_index=job_index,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        states=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        workers=worker_name,
        custom_query=custom_query,
        verbosity=verbosity,
        wait=wait,
        raise_on_error=raise_on_error,
        interactive=True,
        delete_output=delete_output,
        delete_files=delete_files,
    )


@app_job.command()
def queue_out(
    job_db_id: job_db_id_arg,
    job_index: job_index_arg = None,
) -> None:
    """Print the content of the output files produced by the queue manager."""
    db_id, job_id = get_job_db_ids(job_db_id, job_index)

    cm = get_config_manager()

    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Retrieving info...", total=None)
        jc = get_job_controller()

        job_info = jc.get_job_info(
            job_id=job_id,
            job_index=job_index,
            db_id=db_id,
        )

    if not job_info:
        exit_with_error_msg("No data matching the request")

    remote_dir = job_info.run_dir

    if not remote_dir:
        exit_with_warning_msg("The remote folder has not been created yet")

    out_path = Path(remote_dir, OUT_FNAME)
    err_path = Path(remote_dir, ERR_FNAME)
    out = None
    err = None
    out_error = None
    err_error = None
    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Retrieving files...", total=None)
        worker = cm.get_worker(job_info.worker)
        host = worker.get_host()

        try:
            if worker.get_host().interactive_login:
                with hide_progress(progress):
                    host.connect()
            else:
                host.connect()
            try:
                out_bytes = io.BytesIO()
                host.get(out_path, out_bytes)
                out = out_bytes.getvalue().decode("utf-8")
            except Exception as e:
                out_error = getattr(e, "message", str(e))
            try:
                err_bytes = io.BytesIO()
                host.get(err_path, err_bytes)
                err = err_bytes.getvalue().decode("utf-8")
            except Exception as e:
                err_error = getattr(e, "message", str(e))
        finally:
            try:
                host.close()
            except Exception:
                pass

    if out_error:
        out_console.print(
            f"Error while fetching queue output from {out_path!s}: {out_error}",
            style="red",
        )
    else:
        out_console.print(f"Queue output from {out_path!s}:\n")
        out_console.print(out)

    if err_error:
        out_console.print(
            f"Error while fetching queue error from {err_path!s}: {err_error}",
            style="red",
        )
    else:
        out_console.print(f"Queue error from {err_path!s}:\n")
        out_console.print(err)


@app_job.command()
def report(
    interval: Annotated[
        ReportInterval,
        typer.Argument(
            help="The interval of the trends for the report",
            metavar="INTERVAL",
        ),
    ] = ReportInterval.DAYS,
    num_intervals: Annotated[
        Optional[int],
        typer.Argument(
            help="The number of intervals to consider. Default depends on the interval type",
            metavar="NUM_INTERVALS",
        ),
    ] = None,
):
    """
    Generate a report about the Jobs in the database.
    """
    jc = get_job_controller()

    timezone = datetime.now(tzlocal()).tzname()

    jobs_report = JobsReport.generate_report(
        job_controller=jc,
        interval=interval.value,
        num_intervals=num_intervals,
        timezone=timezone,
    )
    out_console.print(*get_job_report_components(jobs_report))


app_job_set = JFRTyper(
    name="set", help="Commands for setting properties for jobs", no_args_is_help=True
)
app_job.add_typer(app_job_set)


@app_job_set.command()
def worker(
    worker_name: Annotated[
        str,
        typer.Argument(
            help="The name of the worker",
            metavar="WORKER",
        ),
    ],
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    select_worker_name: worker_name_opt = None,
    custom_query: query_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Set the worker for the selected Jobs.
    Only Jobs not in an evolving state (e.g. CHECKED_OUT, UPLOADED, ...).
    """
    jc = get_job_controller()
    execute_multi_jobs_cmd(
        single_cmd=jc.set_job_run_properties,
        multi_cmd=jc.set_job_run_properties,
        job_db_id=None,
        job_index=None,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        states=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        workers=select_worker_name,
        custom_query=custom_query,
        verbosity=verbosity,
        raise_on_error=raise_on_error,
        worker=worker_name,
    )


@app_job_set.command()
def exec_config(
    exec_config_value: Annotated[
        str,
        typer.Argument(
            help="The name of the exec_config",
            metavar="EXEC_CONFIG",
        ),
    ],
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    worker_name: worker_name_opt = None,
    custom_query: query_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Set the exec_config for the selected Jobs.
    Only Jobs not in an evolving state (e.g. CHECKED_OUT, UPLOADED, ...).
    """
    jc = get_job_controller()
    execute_multi_jobs_cmd(
        single_cmd=jc.set_job_run_properties,
        multi_cmd=jc.set_job_run_properties,
        job_db_id=None,
        job_index=None,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        states=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        workers=worker_name,
        custom_query=custom_query,
        verbosity=verbosity,
        raise_on_error=raise_on_error,
        exec_config=exec_config_value,
    )


@app_job_set.command()
def resources(
    resources_value: Annotated[
        str,
        typer.Argument(
            help="The resources to be specified. Can be either a list of"
            "comma separated key=value pairs or a string with the JSON "
            "representation of a dictionary "
            '(e.g \'{"key1.key2": 1, "key3": "test"}\')',
            metavar="RESOURCES",
        ),
    ],
    replace: Annotated[
        bool,
        typer.Option(
            "--replace",
            "-r",
            help="If present the value will replace entirely those present "
            "instead of updating the DB, otherwise only the selected keys "
            "will be updated",
        ),
    ] = False,
    qresources: Annotated[
        bool,
        typer.Option(
            "--qresources",
            "-qr",
            help="If present the values in `resources_value` will be interpreted as arguments for a QResources object",
        ),
    ] = False,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    worker_name: worker_name_opt = None,
    custom_query: query_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Set the resources for the selected Jobs.
    Only Jobs not in an evolving state (e.g. CHECKED_OUT, UPLOADED, ...)
    """
    resources = str_to_dict(resources_value)

    if qresources:
        resources = QResources(**resources)

    jc = get_job_controller()
    execute_multi_jobs_cmd(
        single_cmd=jc.set_job_run_properties,
        multi_cmd=jc.set_job_run_properties,
        job_db_id=None,
        job_index=None,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        states=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        workers=worker_name,
        custom_query=custom_query,
        verbosity=verbosity,
        raise_on_error=raise_on_error,
        resources=resources,
        update=not replace,
    )


@app_job_set.command()
def priority(
    priority_value: Annotated[
        int,
        typer.Argument(
            help="The priority of the Job",
            metavar="PRIORITY",
        ),
    ],
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    worker_name: worker_name_opt = None,
    custom_query: query_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Set the priority for the selected Jobs.
    Only Jobs not in an evolving state (e.g. CHECKED_OUT, UPLOADED, ...).
    """
    jc = get_job_controller()
    execute_multi_jobs_cmd(
        single_cmd=jc.set_job_run_properties,
        multi_cmd=jc.set_job_run_properties,
        job_db_id=None,
        job_index=None,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        states=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        workers=worker_name,
        custom_query=custom_query,
        verbosity=verbosity,
        raise_on_error=raise_on_error,
        priority=priority_value,
    )


@app_job.command(name="dump", hidden=True)
def job_dump(
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    worker_name: worker_name_opt = None,
    custom_query: query_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    file_path: Annotated[
        str,
        typer.Option(
            "--path",
            "-p",
            help="Path to where the file should be dumped",
        ),
    ] = "jobs_dump.json",
) -> None:
    """Dump to json the documents of the selected Jobs from the DB. For debugging."""
    check_incompatible_opt({"start_date": start_date, "days": days, "hours": hours})
    check_incompatible_opt({"end_date": end_date, "days": days, "hours": hours})
    check_query_incompatibility(
        custom_query,
        [
            job_id,
            db_id,
            flow_id,
            state,
            start_date,
            end_date,
            name,
            metadata,
            days,
            hours,
            worker_name,
        ],
    )

    job_ids_indexes = get_job_ids_indexes(job_id)

    jc = get_job_controller()

    start_date = get_start_date(start_date, days, hours)

    with loading_spinner():
        if custom_query:
            jobs_doc = jc.get_jobs_doc_query(
                query=custom_query,
            )
        else:
            jobs_doc = jc.get_jobs_doc(
                job_ids=job_ids_indexes,
                db_ids=db_id,
                flow_ids=flow_id,
                states=state,
                start_date=start_date,
                end_date=end_date,
                name=name,
                metadata=metadata,
                workers=worker_name,
            )
        if jobs_doc:
            dumpfn(jsanitize(jobs_doc, strict=True, enum_values=True), file_path)

    if not jobs_doc:
        exit_with_error_msg("No data matching the request")


@app_job.command()
def output(
    job_db_id: job_db_id_arg,
    job_index: job_index_arg = None,
    file_path: Annotated[
        OptionalStr,
        typer.Option(
            "--path",
            "-p",
            help="If defined, the output will be dumped to this file based on the extension (json or yaml)",
        ),
    ] = None,
    load: Annotated[
        bool,
        typer.Option(
            "--load",
            "-l",
            help="If enabled all the data from additional stores are also loaded ",
        ),
    ] = False,
) -> None:
    """Fetch the output of a Job from the output Store."""
    db_id, job_id = get_job_db_ids(job_db_id, job_index)

    with loading_spinner():
        jc = get_job_controller()

        if db_id:
            job_info = jc.get_job_info(
                job_id=job_id,
                job_index=job_index,
                db_id=db_id,
            )
            if job_info:
                job_id = job_info.uuid
                job_index = job_info.index

        job_output = None
        if job_id:
            job_output = jc.jobstore.get_output(job_id, job_index or "last", load=load)

    if not job_output:
        exit_with_error_msg("No data matching the request")

    if file_path:
        dumpfn(job_output, file_path)
    else:
        pprint(job_output)


app_job_files = JFRTyper(
    name="files",
    help="Commands for managing the files associated to a job",
    no_args_is_help=True,
)
app_job.add_typer(app_job_files)


@app_job_files.command(name="ls")
def files_list(
    job_db_id: job_db_id_arg,
    job_index: job_index_arg = None,
) -> None:
    """List of files in the run_dir of the selected Job."""
    db_id, job_id = get_job_db_ids(job_db_id, job_index)

    cm = get_config_manager()

    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Retrieving info...", total=None)
        jc = get_job_controller()

        job_info = jc.get_job_info(
            job_id=job_id,
            job_index=job_index,
            db_id=db_id,
        )

    if not job_info:
        exit_with_error_msg("No data matching the request")

    remote_dir = job_info.run_dir

    if not remote_dir:
        exit_with_warning_msg("The remote folder has not been created yet")

    remote_files = None
    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Retrieving information...", total=None)
        worker = cm.get_worker(job_info.worker)
        host = worker.get_host()

        try:
            host.connect()
            try:
                remote_files = host.listdir(remote_dir)
            except Exception as e:
                raise RuntimeError(
                    f"Error while fetching the list of files from {remote_dir!s}: {getattr(e, 'message', str(e))}"
                ) from e
        finally:
            try:
                host.close()
            except Exception:
                pass

    out_console.print(f"List of files in {remote_dir}:")
    for fn in remote_files:
        out_console.print(fn)


@app_job_files.command(name="get")
def files_get(
    job_db_id: job_db_id_arg,
    filenames: Annotated[
        list[str],
        typer.Argument(
            help="A list of file names to be retrieved from the job run dir",
            metavar="FILE_NAMES",
        ),
    ],
    job_index: job_index_opt = None,
    path: Annotated[
        OptionalStr,
        typer.Option(
            "--path",
            "-p",
            help="If defined, the files will be copied to this path. Otherwise in the local folder.",
        ),
    ] = None,
) -> None:
    """
    Retrieve files from the Job's execution folder.
    """
    db_id, job_id = get_job_db_ids(job_db_id, job_index)

    cm = get_config_manager()

    if not path:
        path = "."
    save_path = Path(path)
    if not save_path.is_dir():
        raise typer.BadParameter("The path is not an existing directory")

    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Retrieving info...", total=None)
        jc = get_job_controller()

        job_info = jc.get_job_info(
            job_id=job_id,
            job_index=job_index,
            db_id=db_id,
        )

    if not job_info:
        exit_with_error_msg("No data matching the request")

    remote_dir = job_info.run_dir

    if not remote_dir:
        exit_with_warning_msg("The remote folder has not been created yet")

    with loading_spinner(processing=False) as progress:
        task_id = progress.add_task(description="Retrieving files...", total=None)
        worker = cm.get_worker(job_info.worker)
        host = worker.get_host()

        file_name: str
        try:
            host.connect()
            for file_name in filenames:
                progress.update(task_id, description=f"Retrieving {file_name}")
                host.get(str(Path(remote_dir) / file_name), str(save_path / file_name))
        except Exception as exc:
            raise RuntimeError(
                f"Error while fetching file {file_name} from {remote_dir!s}: {getattr(exc, 'message', str(exc))}"
            ) from exc
        finally:
            try:
                host.close()
            except Exception:
                pass



from __future__ import annotations

import copy
import logging
import os
import re
import subprocess
import time
import warnings
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from shutil import which
from typing import TYPE_CHECKING, Any

import bson
from maggma.stores.mongolike import MongoStore, MongoURIStore
from monty.io import zopen
from pymongo import ReturnDocument
from pymongo.errors import PyMongoError

from jobflow_remote.utils.data import deep_merge_dict, suuid

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from pymongo.collection import Collection

logger = logging.getLogger(__name__)


class MongoLock:
    """
    Context manager to lock a document in a MongoDB database.

    Main characteristics and functionalities:
        * Lock is acquired by setting a lock_id and lock_time value in the
          locked document.
        * Filter the document to select based on a query and sorting.
          It uses find_one_and_update, thus resulting in a single document locked.
        * Can wait for a lock to be released.
        * Can forcibly break an existing lock
        * Can return the locked document even if the lock could not be acquired
          (useful for determining if a document is locked or no document matches
          the query)
        * Accepts all the arguments that can be passed to find_one_and_update
        * A custom update can be performed on the document when acquiring the lock.
        * Allows to pass properties that will be set in the document at the moment
          of releasing the lock.
        * Lock id value can be customized. If not a randomly generated uuid is used.

    Examples
    --------
    Trying to acquire the lock on a document based on the state
    >>> with MongoLock(collection, {"state": "READY"}) as lock:
    ...     print(lock.locked_document["state"])
    READY

    If lock cannot be acquired (no document matching filter or that
    document is locked) the `lock.locked_document` is None.

    >>> with MongoLock(collection, {"state": "READY"}) as lock:
    ...     print(lock.locked_document)
    None

    Wait for 60 seconds in case the required document is already locked.
    Check the status every 10 seconds. If lock cannot be acquired
    locked_document is None

    >>> with MongoLock(
            collection,
            {"uuid": "5b84228b-d019-47fe-b0a0-564b36aa85ed"},
            sleep=10,
            max_wait=60,
        ) as lock:
    ...     print(lock.locked_document)
    None

    In case lock cannot be acquired expose the locked document

    >>> with MongoLock(
            collection,
            {"uuid": "5b84228b-d019-47fe-b0a0-564b36aa85ed"},
            get_locked_doc=True,
        ) as lock:
    ...     print(lock.locked_document)
    None
    ...     print(lock.unavailable_document['lock_id'])
    8d68404f-c77a-461b-859c-40bb0af1979f

    Set values in the document upon lock release

    >>> with MongoLock(collection, {"state": "READY"}) as lock:
    ...     if lock.locked_document:
    ...         # Perform some operations based on the job...
    ...         lock.update_on_release = {"$set": {"state": "CHECKED_OUT"}}

    Delete the locked document upon lock release

    >>> with MongoLock(collection, {"state": "READY"}) as lock:
    ...     if lock.locked_document:
    ...         # decide if the document should be deleted
    ...         lock.delete_on_release = True

    """

    LOCK_KEY = "lock_id"
    LOCK_TIME_KEY = "lock_time"

    def __init__(
        self,
        collection: Collection,
        filter: Mapping[str, Any],  # noqa: A002
        update: Mapping[str, Any] | None = None,
        break_lock: bool = False,
        lock_id: str | None = None,
        sleep: int | None = None,
        max_wait: int = 600,
        projection: Mapping[str, Any] | Iterable[str] | None = None,
        get_locked_doc: bool = False,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        collection
            The MongoDB collection containing the document to lock.
        filter
            A MongoDB query to select the document.
        update
            A dictionary that will be passed to find_one_and_update to update the
            locked document at the moment of acquiring the lock.
        break_lock
            True if the context manager is allowed to forcibly break a lock.
        lock_id
            The is used for the lock in the document. If None a randomly generated
            uuid will be used.
        sleep
            The amount of second to sleep between consecutive checks while waiting
            for a lock to be released.
        max_wait
            The amount of seconds to wait for a lock to be released.
        projection
            The projection passed to the find_one_and_update that locks the document.
        get_locked_doc
            If True, if the lock cannot be acquired because the document matching
            the filter is already locked, the locked document will be fetched and
            set in the unavailable_document attribute.
        kwargs
            All the other args are passed to find_one_and_update.
        """
        self.collection = collection
        self.filter = filter or {}
        self.update = update
        self.break_lock = break_lock
        self.locked_document = None
        self.unavailable_document = None
        self.lock_id = lock_id or suuid()
        self.kwargs = kwargs
        self._update_on_release: dict | list = {}
        self._delete_on_release: bool = False
        self.sleep = sleep
        self.max_wait = max_wait
        self.projection = projection
        self.get_locked_doc = get_locked_doc

    @property
    def update_on_release(self) -> dict | list:
        """
        The update_on_release value.

        Returns:
            dict | list: The value of update_on_release.
        """
        return self._update_on_release

    @update_on_release.setter
    def update_on_release(self, value: dict | list):
        """
        Set the value of update_on_release.

        If set its value will be used to update the document with additional
        properties upon lock release.

        For example:
        lock.update_on_release = {"$set": {"state": "CHECKED_OUT"}}

        Cannot be set together with delete_on_release.

        Parameters
        ----------
        value : dict | list
            The value to set for update_on_release.
        """
        if self.delete_on_release:
            raise ValueError(
                "delete_on_release and update_on_release cannot be set simultaneously"
            )
        self._update_on_release = value

    @property
    def delete_on_release(self) -> bool:
        """
        The delete_on_release property.

        Returns:
            bool: Whether the document will be deleted upon lock release.
        """
        return self._delete_on_release

    @delete_on_release.setter
    def delete_on_release(self, value: bool):
        """
        Set the value of delete_on_release.

        If True the document will be deleted upon lock release.

        For example:
        lock.delete_on_release = True

        Cannot be set together with update_on_release.

        Parameters
        ----------
        value : bool
            The value to set for delete_on_release.
        """
        if self.update_on_release:
            raise ValueError(
                "delete_on_release and update_on_release cannot be set simultaneously"
            )
        self._delete_on_release = value

    @property
    def is_locked(self) -> bool:
        """Return whether the document was locked before trying to acquire the lock.

        Notes
        -----
        This method should be used only inside the "with" context.
        """
        return self.locked_document is None

    @classmethod
    def get_lock_time(cls, d: dict):
        """Get the time the document was locked on a dictionary."""
        return d.get(cls.LOCK_TIME_KEY)

    @classmethod
    def get_lock_id(cls, d: dict):
        """Get the lock id on a dictionary."""
        return d.get(cls.LOCK_KEY)

    def acquire(self) -> None:
        """Acquire the lock."""
        # Set the lock expiration time
        now = datetime.utcnow()
        db_filter = copy.deepcopy(dict(self.filter))

        projection = self.projection
        # if projecting always get the lock as well
        if projection:
            projection = list(projection)
            projection.extend([self.LOCK_KEY, self.lock_id])

        # Modify the filter if the document should not be fetched if
        # the lock cannot be acquired. Otherwise, keep the original filter.
        if not self.break_lock and not self.sleep and not self.get_locked_doc:
            db_filter.update({self.LOCK_KEY: None})

        # Prepare the update to be performed when acquiring the lock.
        # A combination of the input update and the setting of the lock.
        lock_set = {self.LOCK_KEY: self.lock_id, self.LOCK_TIME_KEY: now}
        update: dict[str, dict] = defaultdict(dict)
        if self.update:
            update.update(copy.deepcopy(self.update))

        update["$set"].update(lock_set)

        # If the document should be fetched even if the lock could not be acquired
        # the updates should be made conditional.
        # Note: sleep needs to fetch the document, otherwise it is impossible to
        # determine if the filter did not return any document or if the document
        # was locked.
        if (self.sleep or self.get_locked_doc) and not self.break_lock:
            for operation, dict_vals in update.items():
                for k, v in dict_vals.items():
                    cond = {
                        "$cond": {
                            "if": {"$gt": [f"${self.LOCK_KEY}", None]},
                            "then": f"${k}",
                            "else": v,
                        }
                    }
                    update[operation][k] = cond
            update = [dict(update)]  # type: ignore[assignment]

        # Try to acquire the lock by updating the document with a unique identifier
        # and the lock expiration time
        logger.debug(f"try acquiring lock with filter: {db_filter}")
        t0 = time.time()
        while True:
            result = self.collection.find_one_and_update(
                db_filter,
                update,
                upsert=False,
                return_document=ReturnDocument.AFTER,
                **self.kwargs,
            )

            if result:
                lock_acquired = self.get_lock_id(result) == self.lock_id
                if lock_acquired:
                    self.locked_document = result
                    break
                # if the lock could not be acquired optionally sleep or
                # exit if waited for enough time.
                if self.sleep and (time.time() - t0) < self.max_wait:
                    logger.debug("sleeping")
                    time.sleep(self.sleep)
                else:
                    self.unavailable_document = result
                    break
            else:
                # If no document the conditions could not be met.
                # Either the requested filter does not find match a document
                # or those fitting are locked.
                break

    def release(self, exc_type, exc_val, exc_tb) -> None:
        """Release the lock."""

        # TODO if failed to release the lock maybe retry before failing
        if self.locked_document is None:
            return

        # Release the lock by removing the unique identifier and lock expiration time
        update: list | dict = {"$set": {self.LOCK_KEY: None, self.LOCK_TIME_KEY: None}}
        # TODO maybe set on release only if no exception was raised?
        if self.update_on_release:
            # if an exception raised inside the context manager do not update the document
            if exc_type is not None:
                logger.warning(
                    f"A {type(exc_type)} exception was raised while the document was locked. "
                    f"The update_on_release {self.update_on_release} will not be applied."
                )
            elif isinstance(self.update_on_release, list):
                update = [update, *self.update_on_release]
            else:
                update = deep_merge_dict(update, self.update_on_release)
        logger.debug(f"release lock with update: {update}")

        # if an exception raised inside the context manager do not delete the document
        if self.delete_on_release and exc_type is None:
            result = self.collection.delete_one(
                {"_id": self.locked_document["_id"], self.LOCK_KEY: self.lock_id}
            )
            if result.deleted_count == 0:
                raise RuntimeError("Could not delete the locked document upon release")
        else:
            if self.delete_on_release and exc_type is not None:
                logger.warning(
                    f"A {type(exc_type)} exception was raised while the document was locked. "
                    f"The document will not be deleted, as instead requested by delete_on_release."
                )
            result = self.collection.update_one(
                {"_id": self.locked_document["_id"], self.LOCK_KEY: self.lock_id},
                update,
                upsert=False,
            )

            # Check if the lock was successfully released
            if result.modified_count == 0:
                msg = (
                    f"Could not release lock for document {self.locked_document['_id']}"
                )
                warnings.warn(msg, stacklevel=2)

        self.locked_document = None

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.locked_document:
            self.release(exc_type, exc_val, exc_tb)


class LockedDocumentError(Exception):
    """Exception to signal a problem when locking the document."""


class RunnerLockedError(LockedDocumentError):
    """Exception to signal a problem when locking a Runner document."""

    @classmethod
    def from_runner_doc(cls, doc: dict, additional_msg: str | None = None):
        lock_id = doc[MongoLock.LOCK_KEY]
        lock_date = doc[MongoLock.LOCK_TIME_KEY]
        date_str = lock_date.isoformat(timespec="seconds") if lock_date else None
        msg = f"Runner document is locked with lock_id {lock_id} since {date_str} UTC."
        if additional_msg:
            msg += " " + additional_msg
        return cls(msg)


class JobLockedError(LockedDocumentError):
    """Exception to signal a problem when locking a Job document."""

    @classmethod
    def from_job_doc(cls, doc: dict, additional_msg: str | None = None):
        lock_id = doc[MongoLock.LOCK_KEY]
        lock_date = doc[MongoLock.LOCK_TIME_KEY]
        date_str = lock_date.isoformat(timespec="seconds") if lock_date else None
        msg = f"Job with db_id {doc['db_id']} is locked with lock_id {lock_id} since {date_str} UTC."
        if additional_msg:
            msg += " " + additional_msg
        return cls(msg)


class FlowLockedError(LockedDocumentError):
    """Exception to signal a problem when locking a Flow document."""

    @classmethod
    def from_flow_doc(cls, doc: dict, additional_msg: str | None = None):
        lock_id = doc[MongoLock.LOCK_KEY]
        lock_date = doc[MongoLock.LOCK_TIME_KEY]
        date_str = lock_date.isoformat(timespec="seconds") if lock_date else None
        msg = f"Flow with uuid {doc['uuid']} is locked with lock_id {lock_id} since {date_str} UTC."
        if additional_msg:
            msg += " " + additional_msg
        return cls(msg)


class MissingDocumentError(Exception):
    """Exception to signal that a document is missing from the DB"""


def mongo_operation(
    store: MongoStore | MongoURIStore,
    file_path: str | Path,
    operation: str,
    collection: str | None,
    mongo_bin_path: str | None = None,
    compress: bool = False,
) -> tuple[str, str]:
    """
    Execute a mongo operation (mongodump or mongorestore) on a given store.

    Parameters
    ----------
    store
        The store containing the data to be backed up.
    file_path
        The path of the folder where the backup files are located.
    operation
        The mongo operation to perform. Can be either 'mongodump' or 'mongorestore'.
    collection
        The name of the collection to be backed up. If None the collection defined in
        the store will be used.
    mongo_bin_path
        The path to the folder containing the mongo executable. If None, the executable is
        searched in the PATH.
    compress
        If True, the backup files are compressed.

    Returns
    -------
    tuple[str, str]
        The stdout and stderr of the executed command.
    """
    if operation not in ("mongodump", "mongorestore"):
        raise ValueError(f"Operation {operation} not supported")

    if mongo_bin_path:
        operation = os.path.join(mongo_bin_path, operation)

    if not which(operation):
        raise RuntimeError(
            f"It looks like the command {operation} is not available. Check the path or consider "
            f"installing the mongodb database tools. Alternatively use the pure python implementation."
        )

    # here is not checked with isinstance(), because the current other subclasses will fail
    if type(store) is MongoURIStore:
        cmd = [
            operation,
            "--uri",
            store.uri,
        ]
    elif type(store) is MongoStore:
        cmd = [
            operation,
            "--host",
            store.host,
            "--port",
            str(store.port),
        ]
        if store.username and store.password:
            cmd.extend(["--username", store.username, "--password", store.password])
            if store.auth_source:
                cmd.extend(["--authenticationDatabase", store.auth_source])
        elif store.username or store.password:
            raise ValueError(
                "To use the mongotools the username and password in the queue store should be "
                "either both present or both absent."
            )
    else:
        raise ValueError(
            f"Unsupported store type {type(store).__name__}. Consider using the python version."
        )

    # check this afterwards, since other stores may not have ssh_tunnel attribute
    if store.ssh_tunnel is not None:
        raise NotImplementedError(
            "SSH tunnel is not supported. Consider using the python version."
        )

    cmd.extend(["--db", store.database, "--collection", collection])

    if compress:
        cmd.append("--gzip")

    if operation.endswith("mongodump"):
        cmd.extend(["--out", file_path])
    elif operation.endswith("mongorestore"):
        cmd.append(file_path)

    str_cmd = " ".join(str(s) for s in cmd)
    result = subprocess.run(
        str_cmd, check=True, capture_output=True, text=True, shell=True
    )
    logger.debug(
        f"output during execution of '{str_cmd}'. Stdout: {result.stdout}. Stderr: {result.stderr}"
    )
    return result.stdout, result.stderr


def mongodump_from_store(
    store: MongoStore | MongoURIStore,
    output_path: str | Path,
    collection: str | None = None,
    mongo_bin_path: str | None = None,
    compress: bool = False,
) -> int:
    """
    Use mongodump to dump a collection from a MongoDB store to a file.

    Parameters
    ----------
    store
        The store containing the data to be backed up.
    output_path
        The path of the folder where the backup files are located.
    collection
        The name of the collection to be backed up. If None the collection defined in
        the store will be used.
    mongo_bin_path
        The path to the folder containing the mongo executable. If None, the executable is
        searched in the PATH.
    compress
        If True, the backup files are compressed.

    Returns
    -------
    int
        The number of documents dumped.
    """
    stdout, stderr = mongo_operation(
        store=store,
        collection=collection,
        file_path=output_path,
        operation="mongodump",
        mongo_bin_path=mongo_bin_path,
        compress=compress,
    )

    # Extract the number of documents dumped
    match = re.search(r"done dumping .*\((\d+) documents*\)", stderr)
    return int(match.group(1)) if match else 0


def mongorestore_to_store(
    store: MongoStore | MongoURIStore,
    input_file: str | Path,
    collection: str | None = None,
    mongo_bin_path: str | None = None,
    compress: bool | None = None,
) -> None:
    """
    Restore a collection from a BSON file using mongorestore.

    Parameters
    ----------
    store
        The store where the data should be restored.
    input_file
        The path of the BSON file containing the data to be restored.
    collection
        The name of the collection to be backed up. If None the collection defined in
        the store will be used.
    mongo_bin_path
        The path to the folder containing the mongo executable. If None, the executable is
        searched in the PATH.
    compress
        If True, the input file is expected to be compressed. If None it will be determined
        based on the file extension.
    """
    if compress is None:
        compress = Path(input_file).name.endswith(".gz")
    mongo_operation(
        store=store,
        collection=collection,
        file_path=input_file,
        operation="mongorestore",
        mongo_bin_path=mongo_bin_path,
        compress=compress,
    )


def pymongo_dump(
    collection: Collection, output_path: str | Path, compress: bool = False
) -> int:
    """
    Dump the contents of a PyMongo collection to a BSON file.

    Parameters
    ----------
    collection
        The PyMongo collection to be dumped.
    output_path
        The path of the folder where the BSON file should be saved.
    compress : bool
        If True, the output file is compressed.

    Returns
    -------
    int
        The number of documents dumped.
    """
    dir_path = Path(output_path) / collection.database.name
    dir_path.mkdir(exist_ok=True, parents=True)
    # add the db name for consistency with mongodump and use the collection name as file name
    file_name = f"{collection.name}.bson"
    if compress:
        file_name += ".gz"
    with zopen(dir_path / file_name, "wb") as f:
        num_documents = 0
        for doc in collection.find():
            f.write(bson.BSON.encode(doc))
            num_documents += 1

    return num_documents


def pymongo_restore(collection: Collection, input_file: str | Path) -> None:
    """
    Restore the contents of a BSON file to a PyMongo collection.

    Parameters
    ----------
    collection
        The PyMongo collection to be dumped.
    input_file
        The path of the BSON file used to restore the data.
    """
    try:
        with zopen(input_file, "rb") as f:
            collection.insert_many(bson.decode_all(f.read()))
    except PyMongoError as e:
        raise RuntimeError(f"Error during PyMongo restore: {e!s}") from e
    except OSError as e:
        raise RuntimeError(f"Error reading from file: {e!s}") from e



from typing import Annotated, Optional

import typer

from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.utils import (
    check_stopped_runner,
    get_job_controller,
    loading_spinner,
    out_console,
)

app_backup = JFRTyper(
    name="backup",
    help="Commands for handling backup of the database",
    no_args_is_help=True,
)
app.add_typer(app_backup)


@app_backup.command()
def create(
    backup_dir: Annotated[
        str,
        typer.Argument(
            help="The path of the directory where the output files will be saved",
            metavar="BACKUP_DIR",
        ),
    ] = ".",
    compress: Annotated[
        bool,
        typer.Option(
            "--compress",
            "-c",
            help="Compress the output files",
        ),
    ] = False,
    mongo_path: Annotated[
        Optional[str],
        typer.Option(
            "--mongo-path",
            "-m",
            help=(
                "The path to a folder containing the mongodump executable, if not present in the PATH"
            ),
        ),
    ] = None,
    python: Annotated[
        bool,
        typer.Option(
            "--python",
            "-p",
            help="If selected a python implementation will be used to create a backup. "
            "WARNING: In this case metadata of the collections will not be saved",
        ),
    ] = False,
) -> None:
    """
    Create a backup of the queue database using either mongodump or a python implementation.
    The mongodump version is faster and stores metadata, but requires the mongodump executable
    and may not support all the connection options defined in the project configuration.
    """
    check_stopped_runner(error=True)

    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Creating backup...", total=None)
        jc = get_job_controller()
        n_docs = jc.backup_dump(
            dir_path=backup_dir,
            mongo_bin_path=mongo_path,
            compress=compress,
            python=python,
        )

    out_console.print("Backup created for ")
    for collection in sorted(n_docs):
        out_console.print(f"\t{collection} collection: {n_docs[collection]} documents")
    out_console.print("Verify that these numbers fit the expected values.")


@app_backup.command()
def restore(
    backup_dir: Annotated[
        str,
        typer.Argument(
            help="The path of the directory where the backup files are located",
            metavar="BACKUP_DIR",
        ),
    ] = ".",
    mongo_path: Annotated[
        Optional[str],
        typer.Option(
            "--mongo-path",
            "-m",
            help=(
                "The path to a folder containing the mongorestore executable, if not present in the PATH"
            ),
        ),
    ] = None,
    python: Annotated[
        bool,
        typer.Option(
            "--python",
            "-p",
            help="If selected a python implementation will be used to restore a backup. "
            "WARNING: In this case metadata of the collections will not be restored",
        ),
    ] = False,
) -> None:
    """
    Recreate the queue database from a previous backup using either mongorestore or a python implementation.
    """
    check_stopped_runner(error=True)

    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Restoring backup...", total=None)
        jc = get_job_controller()
        jc.backup_restore(
            dir_path=backup_dir,
            mongo_bin_path=mongo_path,
            python=python,
        )

    out_console.print("Backup restored")
    if python:
        out_console.print(
            "Python version does not restore indexes in the DB. It is advisable to run "
            "'jf admin index rebuild' to create them."
        )



from __future__ import annotations

from enum import Enum


class JobState(Enum):
    """States of a Job."""

    WAITING = "WAITING"
    READY = "READY"
    CHECKED_OUT = "CHECKED_OUT"  # TODO should it be RESERVED?
    UPLOADED = "UPLOADED"
    SUBMITTED = "SUBMITTED"
    RUNNING = "RUNNING"
    TERMINATED = "TERMINATED"
    DOWNLOADED = "DOWNLOADED"
    REMOTE_ERROR = "REMOTE_ERROR"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    PAUSED = "PAUSED"
    STOPPED = "STOPPED"
    USER_STOPPED = "USER_STOPPED"
    BATCH_SUBMITTED = "BATCH_SUBMITTED"
    BATCH_RUNNING = "BATCH_RUNNING"

    @property
    def short_value(self) -> str:
        return short_state_mapping[self]


short_state_mapping = {
    JobState.WAITING: "W",
    JobState.READY: "R",
    JobState.CHECKED_OUT: "CE",
    JobState.UPLOADED: "U",
    JobState.SUBMITTED: "SU",
    JobState.RUNNING: "RU",
    JobState.TERMINATED: "T",
    JobState.DOWNLOADED: "D",
    JobState.REMOTE_ERROR: "RERR",
    JobState.COMPLETED: "C",
    JobState.FAILED: "F",
    JobState.PAUSED: "P",
    JobState.STOPPED: "ST",
    JobState.USER_STOPPED: "CA",
    JobState.BATCH_SUBMITTED: "BS",
    JobState.BATCH_RUNNING: "BR",
}


PAUSABLE_STATES = [
    JobState.READY,
    JobState.WAITING,
]

PAUSABLE_STATES_V = [s.value for s in PAUSABLE_STATES]

RUNNING_STATES = [
    JobState.CHECKED_OUT,
    JobState.UPLOADED,
    JobState.SUBMITTED,
    JobState.RUNNING,
    JobState.TERMINATED,
    JobState.DOWNLOADED,
]

RUNNING_STATES_V = [s.value for s in RUNNING_STATES]

RESETTABLE_STATES = RUNNING_STATES

RESETTABLE_STATES_V = RUNNING_STATES_V

DELETABLE_STATES = [
    s for s in JobState if s not in [JobState.BATCH_SUBMITTED, JobState.BATCH_RUNNING]
]


class FlowState(Enum):
    """States of a Flow."""

    WAITING = "WAITING"
    READY = "READY"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    PAUSED = "PAUSED"
    STOPPED = "STOPPED"

    @classmethod
    def from_jobs_states(
        cls, jobs_states: list[JobState], leaf_states: list[JobState]
    ) -> FlowState:
        """
        Generate the state of the Flow based on the states of the Jobs
        composing it, and in particular the states of the leaf Jobs.

        Parameters
        ----------
        jobs_states
            List of JobStates of all the Jobs in the Flow.
        leaf_states
            List of JobStates of the leaf Jobs in the Flow.

        Returns
        -------
        FlowState
            The state of the Flow.
        """
        if all(js == JobState.WAITING for js in jobs_states):
            return cls.WAITING
        if all(js in (JobState.WAITING, JobState.READY) for js in jobs_states):
            return cls.READY
        # only need to check the leaf states to determine if it is completed,
        # in case some intermediate Job failed but children allow missing
        # references.
        if all(js == JobState.COMPLETED for js in leaf_states):
            return cls.COMPLETED
        # REMOTE_ERROR state does not lead to a failed Flow. Two main reasons:
        # 1) it might be a temporary problem and not a final failure of the Flow
        # 2) Changing the state of the flow would require locking the Flow
        #    when applying the change in the remote state.
        if any(js == JobState.FAILED for js in jobs_states):
            return cls.FAILED
        if any(js in (JobState.STOPPED, JobState.USER_STOPPED) for js in jobs_states):
            return cls.STOPPED
        if any(js == JobState.PAUSED for js in jobs_states):
            return cls.PAUSED
        return cls.RUNNING



from jobflow_remote.config.base import (
    ConfigError,
    LocalWorker,
    Project,
    RemoteWorker,
    RunnerOptions,
)
from jobflow_remote.config.manager import ConfigManager, ProjectData

__all__ = (
    "ConfigError",
    "ConfigManager",
    "LocalWorker",
    "Project",
    "ProjectData",
    "RemoteWorker",
    "RunnerOptions",
)



from random import randint
from typing import NoReturn

from jobflow import Flow, Response, job


@job
def add(a, b):
    """A Job adding to numbers."""
    return a + b


@job
def add_many(*to_sum):
    """A job adding numbers."""
    return sum(to_sum)


@job
def sleep(s):
    """A Job sleeping."""
    import time

    time.sleep(s)
    return s


@job
def add_raise(a, b) -> NoReturn:
    """A Job raising a RuntimeError."""
    raise RuntimeError("An error for a and b")


@job
def make_list(a, length=None):
    """A Job generating a list of numbers."""
    if not length:
        length = randint(2, 5)
    return [a] * length


@job
def add_distributed(list_a):
    """A Job generating a new Flow to add a list of numbers."""
    jobs = [add(val, 1) for val in list_a]
    flow = Flow(jobs)
    return Response(replace=flow)


@job
def value(n):
    """A Job returning the input value."""
    return n


@job
def conditional_sum_replace(numbers, min_n=10):
    """A Job creating a replace and adding number until a value is reached."""
    tot = sum(numbers)
    if tot < min_n:
        j1 = value(tot)
        j2 = value(5)
        c = conditional_sum_replace([j1.output, j2.output], min_n=min_n)
        return Response(replace=Flow([j1, j2, c], output=c.output))
    return tot



from typing import TYPE_CHECKING, Annotated

import typer
from rich.prompt import Confirm
from rich.text import Text

from jobflow_remote.cli.formatting import get_exec_config_table, get_worker_table
from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.types import (
    force_opt,
    serialize_file_format_opt,
    tree_opt,
    verbosity_opt,
)
from jobflow_remote.cli.utils import (
    SerializeFileFormat,
    check_incompatible_opt,
    exit_with_error_msg,
    exit_with_warning_msg,
    get_config_manager,
    hide_progress,
    loading_spinner,
    out_console,
    print_success_msg,
)
from jobflow_remote.config import ConfigError, ConfigManager
from jobflow_remote.config.helper import (
    check_jobstore,
    check_queue_store,
    check_worker,
    generate_dummy_project,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

app_project = JFRTyper(
    name="project",
    help="Commands concerning the project definition",
    # no_args_is_help=True,
)
app.add_typer(app_project)


@app_project.command(name="list")
def list_projects(
    warn: Annotated[
        bool,
        typer.Option(
            "--warn",
            "-w",
            help="Print the warning for the files that could not be parsed",
        ),
    ] = False,
) -> None:
    """List of available projects."""
    cm = ConfigManager(warn=warn)

    project_name = None
    try:
        project_data = cm.get_project_data()
        project_name = project_data.project.name
    except ConfigError:
        pass

    full_project_list = cm.project_names_from_files()

    if not full_project_list:
        exit_with_warning_msg(f"No project available in {cm.projects_folder}")

    out_console.print(f"List of projects in {cm.projects_folder}")
    for pn in sorted(full_project_list):
        out_console.print(f" - {pn}", style="green" if pn == project_name else None)

    not_parsed_projects = set(full_project_list).difference(cm.projects_data)
    if not_parsed_projects:
        out_console.print(
            "The following project names exist in files in the project folder, "
            "but could not properly parsed as projects: "
            f"{', '.join(not_parsed_projects)}.",
            style="yellow",
        )
        from jobflow_remote import SETTINGS

        if SETTINGS.cli_suggestions:
            out_console.print(
                "Run the command with -w option to see the parsing errors",
                style="yellow",
            )


@app_project.callback(invoke_without_command=True)
def current_project(
    ctx: typer.Context,
    print_tree: tree_opt = False,  # If selected will print the tree of the CLI and exit
) -> None:
    """Print the list of the project currently selected."""
    # only run if no other subcommand is executed
    if ctx.invoked_subcommand is None:
        out_console.print("Run 'jf project -h' to get the list of available commands")


@app_project.command()
def generate(
    name: Annotated[str, typer.Argument(help="Name of the project")],
    file_format: serialize_file_format_opt = SerializeFileFormat.YAML,
    full: Annotated[
        bool,
        typer.Option(
            "--full",
            help="Generate a configuration file with all the fields and more elements",
        ),
    ] = False,
) -> None:
    """Generate a project configuration file with dummy elements to be edited manually."""
    cm = ConfigManager(exclude_unset=not full)
    if name in cm.projects_data:
        exit_with_error_msg(f"Project with name {name} already exists")

    filepath = cm.projects_folder / f"{name}.{file_format.value}"
    if filepath.exists():
        exit_with_error_msg(
            f"Project with name {name} does not exist, but file {filepath!s} does and will not be overwritten"
        )

    project = generate_dummy_project(name=name, full=full)
    cm.create_project(project, ext=file_format.value)
    print_success_msg(f"Configuration file for project {name} created in {filepath!s}")


@app_project.command()
def check(
    jobstore: Annotated[
        bool,
        typer.Option(
            "--jobstore",
            "-js",
            help="Only check the jobstore connection",
        ),
    ] = False,
    queue: Annotated[
        bool,
        typer.Option(
            "--queue",
            "-q",
            help="Only check the queue connection",
        ),
    ] = False,
    worker: Annotated[
        str,
        typer.Option(
            "--worker",
            "-w",
            help="Only check the connection for the selected worker",
        ),
    ] = None,
    print_errors: Annotated[
        bool,
        typer.Option(
            "--errors",
            "-e",
            help="Print the errors at the end of the checks",
        ),
    ] = False,
    full: Annotated[
        bool,
        typer.Option(
            "--full",
            "-f",
            help="Perform a full check",
        ),
    ] = False,
) -> None:
    """Check that the connection to the different elements of the projects are working."""
    check_incompatible_opt({"jobstore": jobstore, "queue": queue, "worker": worker})

    cm = get_config_manager()
    project = cm.get_project()

    check_all = all(not v for v in (jobstore, worker, queue))

    workers_to_test: Iterable[str] = []
    if check_all:
        workers_to_test = project.workers
    elif worker:
        if worker not in project.workers:
            exit_with_error_msg(
                f"Worker {worker} does not exists in project {project.name}"
            )
        workers_to_test = [worker]

    tick = "[bold green]✓[/] "
    tick_warn = "[bold yellow]✓[/] "
    cross = "[bold red]x[/] "
    errors = []
    with loading_spinner(processing=False) as progress:
        task_id = progress.add_task("Checking")
        for worker_name in workers_to_test:
            progress.update(task_id, description=f"Checking worker {worker_name}")
            worker_to_test = project.workers[worker_name]
            if worker_to_test.get_host().interactive_login:
                with hide_progress(progress):
                    err, worker_warn = check_worker(worker_to_test, full_check=full)
            else:
                err, worker_warn = check_worker(worker_to_test, full_check=full)
            header = tick
            # At the moment the check_worker should return either an error or a
            # warning. The code below also deals with the case where both are
            # returned in the future.
            if worker_warn:
                errors.append((f"Worker {worker_name} warning ", worker_warn))
                header = tick_warn
            if err:
                errors.append((f"Worker {worker_name} ", err))
                header = cross
            progress.print(Text.from_markup(header + f"Worker {worker_name}"))

        if check_all or jobstore:
            progress.update(task_id, description="Checking jobstore")
            err = check_jobstore(project.get_jobstore())
            header = tick
            if err:
                errors.append(("Jobstore", err))
                header = cross
            progress.print(Text.from_markup(header + "Jobstore"))

        if check_all or queue:
            progress.update(task_id, description="Checking queue store")
            err = check_queue_store(project.get_queue_store())
            header = tick
            if err:
                errors.append(("Queue store", err))
                header = cross
            progress.print(Text.from_markup(header + "Queue store"))

    if print_errors and errors:
        out_console.print("Errors:", style="red bold")
        for e in errors:
            out_console.print(e[0], style="bold")
            out_console.print(e[1])


@app_project.command()
def remove(
    name: Annotated[str, typer.Argument(help="Name of the project")],
    keep_folders: Annotated[
        bool,
        typer.Option(
            "--keep-folders",
            "-k",
            help="Project related folders are not deleted",
        ),
    ] = False,
    force: force_opt = False,
) -> None:
    """Remove a project from the projects' folder, including the related folders."""
    cm = get_config_manager()

    if name not in cm.projects_data:
        exit_with_warning_msg(f"Project {name} does not exist")

    p = cm.get_project(name)

    if not keep_folders and not force:
        msg = f"This will delete also the folders:\n\t{p.base_dir}\n\t{p.log_dir}\n\t{p.tmp_dir}\n\t{p.daemon_dir}\nProceed anyway?"
        if not Confirm.ask(msg):
            raise typer.Exit(0)

    with loading_spinner(processing=False) as progress:
        progress.add_task("Deleting project")
        cm.remove_project(project_name=name, remove_folders=not keep_folders)


#####################################
# Exec config app
#####################################


app_exec_config = JFRTyper(
    name="exec_config",
    help="Commands concerning the Execution configurations",
    no_args_is_help=True,
)
app_project.add_typer(app_exec_config)


@app_exec_config.command(name="list")
def list_exec_config(
    verbosity: verbosity_opt = 0,
) -> None:
    """
    The list of defined Execution configurations
    """
    cm = get_config_manager()
    project = cm.get_project()
    table = get_exec_config_table(project.exec_config, verbosity)
    out_console.print(table)


#####################################
# Worker app
#####################################


app_worker = JFRTyper(
    name="worker",
    help="Commands concerning the workers",
    no_args_is_help=True,
)
app_project.add_typer(app_worker)


@app_worker.command(name="list")
def list_worker(
    verbosity: verbosity_opt = 0,
) -> None:
    """
    The list of defined workers
    """
    cm = get_config_manager()
    project = cm.get_project()
    table = get_worker_table(project.workers, verbosity)
    out_console.print(table)



import datetime
import os
from typing import Annotated

import typer
from rich.prompt import Confirm
from rich.scope import render_scope
from rich.table import Table
from rich.text import Text

from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.types import (
    break_lock_opt,
    force_opt,
    log_level_opt,
    verbosity_opt,
)
from jobflow_remote.cli.utils import (
    exit_with_error_msg,
    exit_with_warning_msg,
    fmt_datetime,
    get_config_manager,
    get_job_controller,
    loading_spinner,
    out_console,
)
from jobflow_remote.config.base import LogLevel
from jobflow_remote.jobs.daemon import (
    DaemonError,
    DaemonManager,
    DaemonStatus,
    RunningDaemonError,
)
from jobflow_remote.jobs.runner import Runner
from jobflow_remote.utils.data import convert_utc_time

app_runner = JFRTyper(
    name="runner", help="Commands for handling the Runner", no_args_is_help=True
)
app.add_typer(app_runner)


_running_daemon_error_msg = (
    "\nIf no runner is active on that machine clean the DB with `jf runner reset`"
)


@app_runner.command()
def run(
    log_level: log_level_opt = LogLevel.INFO,
    set_pid: Annotated[
        bool,
        typer.Option(
            "--set-pid",
            "-pid",
            help="Set the runner id to the current process pid",
        ),
    ] = False,
    transfer: Annotated[
        bool,
        typer.Option(
            "--transfer",
            "-t",
            help="Enable the transfer option in the runner",
        ),
    ] = False,
    complete: Annotated[
        bool,
        typer.Option(
            "--complete",
            "-com",
            help="Enable the complete option in the runner",
        ),
    ] = False,
    queue: Annotated[
        bool,
        typer.Option(
            "--queue",
            "-q",
            help="Enable the queue option in the runner",
        ),
    ] = False,
    checkout: Annotated[
        bool,
        typer.Option(
            "--checkout",
            "-cho",
            help="Enable the checkout option in the runner",
        ),
    ] = False,
    connect_interactive: Annotated[
        bool,
        typer.Option(
            "--connect-interactive",
            "-i",
            help="Activate the connection for interactive remote host",
        ),
    ] = False,
) -> None:
    """
    Execute the Runner in the foreground.
    Do NOT execute this to start as a daemon.
    Should be used by the daemon or for testing purposes.
    """
    runner_id = os.getpid() if set_pid else None
    runner = Runner(
        log_level=log_level,
        runner_id=str(runner_id),
        connect_interactive=connect_interactive,
    )
    if not (transfer or complete or queue or checkout):
        transfer = complete = queue = checkout = True

    try:
        runner.run(transfer=transfer, complete=complete, queue=queue, checkout=checkout)
    finally:
        runner.cleanup()


@app_runner.command()
def start(
    transfer: Annotated[
        int,
        typer.Option(
            "--transfer",
            "-t",
            help="The number of processes dedicated to completing jobs",
        ),
    ] = 1,
    complete: Annotated[
        int,
        typer.Option(
            "--complete",
            "-com",
            help="The number of processes dedicated to completing jobs",
        ),
    ] = 1,
    single: Annotated[
        bool,
        typer.Option(
            "--single",
            "-s",
            help="Use a single process for the runner",
        ),
    ] = False,
    log_level: log_level_opt = LogLevel.INFO,
    connect_interactive: Annotated[
        bool,
        typer.Option(
            "--connect-interactive",
            "-i",
            help="Wait for the daemon to start and manually log in the "
            "connection for interactive remote host. Requires --single.",
        ),
    ] = False,
) -> None:
    """Start the Runner as a daemon."""
    # This is not a strict requirement, but for the moment only allow the single
    # process daemon
    if connect_interactive and not single:
        exit_with_error_msg("--connect-interactive option requires --single")
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    with loading_spinner(processing=False) as progress:
        task_id = progress.add_task(description="Starting the daemon...", total=None)
        try:
            dm.start(
                num_procs_transfer=transfer,
                num_procs_complete=complete,
                single=single,
                log_level=log_level.value,
                raise_on_error=True,
                connect_interactive=connect_interactive,
            )
        except RunningDaemonError as e:
            exit_with_error_msg(
                f"Error while starting the daemon: {getattr(e, 'message', e)}{_running_daemon_error_msg}"
            )
        except DaemonError as e:
            exit_with_error_msg(
                f"Error while starting the daemon: {getattr(e, 'message', e)}"
            )
        if connect_interactive:
            progress.update(task_id, description="Waiting for processes to start...")
            try:
                dm.wait_start()
            except DaemonError as e:
                exit_with_error_msg(
                    f"Error while waiting the processes to start: {getattr(e, 'message', e)}"
                )
    if connect_interactive:
        dm.foreground_processes(print_function=out_console.print)


@app_runner.command()
def stop(
    wait: Annotated[
        bool,
        typer.Option(
            "--wait",
            "-w",
            help=(
                "Wait until the daemon has stopped. NOTE: this may take a while if a large file is being transferred!"
            ),
        ),
    ] = False,
) -> None:
    """
    Send a stop signal to the Runner processes.
    Each of the Runner processes will stop when finished the task being executed.
    By default, return immediately.
    """
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Stopping the daemon...", total=None)
        try:
            dm.stop(wait=wait, raise_on_error=True)
        except RunningDaemonError as e:
            exit_with_error_msg(
                f"Error while stopping the daemon: {getattr(e, 'message', e)}{_running_daemon_error_msg}"
            )
        except DaemonError as e:
            exit_with_error_msg(
                f"Error while stopping the daemon: {getattr(e, 'message', e)}"
            )
    from jobflow_remote import SETTINGS

    if not wait and SETTINGS.cli_suggestions:
        out_console.print(
            "The stop signal has been sent to the Runner. Run 'jf runner status' to verify if it stopped",
            style="yellow",
        )


@app_runner.command()
def kill() -> None:
    """
    Send a kill signal to the Runner processes.
    Return immediately, does not wait for processes to be killed.
    """
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Killing the daemon...", total=None)
        try:
            dm.kill(raise_on_error=True)
        except RunningDaemonError as e:
            exit_with_error_msg(
                f"Error while killing the daemon: {getattr(e, 'message', e)}{_running_daemon_error_msg}"
            )
        except DaemonError as e:
            exit_with_error_msg(
                f"Error while killing the daemon: {getattr(e, 'message', e)}"
            )


@app_runner.command()
def shutdown() -> None:
    """
    Shuts down the supervisord process.
    Note that if the daemon is running it will wait for the daemon to stop.
    """
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Shutting down supervisor...", total=None)
        try:
            dm.shut_down(raise_on_error=True)
        except RunningDaemonError as e:
            exit_with_error_msg(
                f"Error while shutting down supervisor: {getattr(e, 'message', e)}{_running_daemon_error_msg}"
            )
        except DaemonError as e:
            exit_with_error_msg(
                f"Error while shutting down supervisor: {getattr(e, 'message', e)}"
            )


@app_runner.command()
def status() -> None:
    """Fetch the status of the daemon runner."""
    from jobflow_remote import SETTINGS

    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    with loading_spinner():
        try:
            current_status = dm.check_status()
        except DaemonError as e:
            exit_with_error_msg(
                f"Error while checking the status of the daemon: {getattr(e, 'message', e)}"
            )
    color = {
        DaemonStatus.STOPPED: "red",
        DaemonStatus.STOPPING: "gold1",
        DaemonStatus.SHUT_DOWN: "red",
        DaemonStatus.PARTIALLY_RUNNING: "gold1",
        DaemonStatus.STARTING: "gold1",
        DaemonStatus.RUNNING: "green",
    }[current_status]
    text = Text()
    text.append("Daemon status: ")
    text.append(current_status.value.lower(), style=color)
    out_console.print(text)
    if current_status == DaemonStatus.PARTIALLY_RUNNING and SETTINGS.cli_suggestions:
        out_console.print(
            f"The {current_status.value.lower()} may be present due to the "
            "runner stopping or signal a problem with one of the processes "
            "of the runner. If the state should be RUNNING, check the detailed"
            " status with the 'info'  command and consider restarting the runner.",
            style="yellow",
        )


@app_runner.command()
def info(verbosity: verbosity_opt = 0) -> None:
    """
    Fetch the information about the process of the daemon.
    Contain the supervisord process and the processes running the Runner.
    """
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    procs_info_dict = None
    try:
        with loading_spinner():
            procs_info_dict = dm.get_processes_info()
    except DaemonError as e:
        exit_with_error_msg(
            f"Error while fetching information from the daemon: {getattr(e, 'message', e)}"
        )
    if not procs_info_dict:
        out_console.print("Daemon is not running", style="gold1")
    else:
        table = Table()
        table.add_column("Process")
        table.add_column("PID")
        table.add_column("State")

        for name, proc_info in procs_info_dict.items():
            table.add_row(name, str(proc_info["pid"]), str(proc_info["statename"]))

        out_console.print(table)

    # add the information about the running_runner according to the DB
    jc = get_job_controller()
    running_runner_doc = jc.get_running_runner()

    # empty line
    out_console.print("")
    if running_runner_doc:
        out_console.print("Data about running runner in the DB:")
        if verbosity == 0:
            running_runner_doc.pop("processes_info", None)
        # convert dates at the first level and for the remote error
        for k, v in running_runner_doc.items():
            if isinstance(v, datetime.datetime):
                running_runner_doc[k] = convert_utc_time(v).strftime(fmt_datetime)
        out_console.print(render_scope(running_runner_doc))
    else:
        out_console.print("No running runner defined in the DB")


@app_runner.command()
def foreground() -> None:
    """Connect to the daemon processes in the foreground."""
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    procs_info_dict = None
    try:
        with loading_spinner():
            procs_info_dict = dm.get_processes_info()
    except DaemonError as e:
        exit_with_error_msg(
            f"Error while fetching information from the daemon: {getattr(e, 'message', e)}"
        )
    if not procs_info_dict:
        exit_with_warning_msg("Daemon is not running")
    dm.foreground_processes(print_function=out_console.print)


@app_runner.command()
def reset(
    force: force_opt = False,
    break_lock: break_lock_opt = False,
) -> None:
    """
    Reset the value of the machine executing the runner from the database.
    Should be executed only if it is certain that the runner is not active on that
    machine anymore.
    """
    jc = get_job_controller()

    running_runner = jc.get_running_runner()
    if running_runner in ("NO_DOCUMENT", None):
        exit_with_warning_msg("No running runner present in the database")
        raise typer.Exit(0)
    if not force:
        text = Text.from_markup(
            "[red]This operation will remove the information about the current "
            "running runner from the database:[/red]\n"
            f"- hostname: {running_runner['hostname']}\n"
            f"- project_name: {running_runner['project_name']}\n"
            f"- start_time: {running_runner['start_time']}\n"
            f"- last_pinged: {running_runner['last_pinged']}\n"
            f"- daemon_dir: {running_runner['daemon_dir']}\n"
            f"- user: {running_runner['user']}\n"
            "[red]Do you want to proceed?[\red]"
        )

        confirmed = Confirm.ask(text, default=False)
        if not confirmed:
            raise typer.Exit(0)

    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Resetting runner information...", total=None)
        jc.clean_running_runner(break_lock=break_lock)

    out_console.print("The running runner document was reset")



"""A series of toy workflows that can be used for testing."""

from typing import Callable, NoReturn, Optional, Union

from jobflow import Job, Response, job


@job
def add(a, b):
    """Adds two numbers together and writes the answer to a file."""
    return a + b


@job
def always_fails() -> NoReturn:
    """A job that always fails."""
    raise RuntimeError("This job failed.")


@job
def write_file(n) -> None:
    with open("results.txt", "w") as f:
        f.write(str(n))


@job
def arithmetic(
    a: Union[float, list[float]],
    b: Union[float, list[float]],
    op: Optional[Callable] = None,
) -> Optional[float]:
    if op:
        return op(a, b)

    return None


@job
def check_env_var() -> str:
    import os

    return os.environ.get("TESTING_ENV_VAR", "unset")


@job(big_data="data")
def add_big(a: float, b: float):
    """Adds two numbers together and inflates the answer
    to a large list and tries to store that within
    the defined store.
    """
    result = a + b
    big_array = [result] * 5_000
    return Response({"data": big_array, "result": a + b})


@job(undefined_store="data")
def add_big_undefined_store(a: float, b: float):
    """Adds two numbers together and writes the answer to an artificially large file
    which is attempted to be stored in a undefined store.
    """
    result = a + b
    return Response({"data": [result] * 5_000, "result": result})


@job
def add_sleep(a, b):
    """Adds two numbers together and sleeps for "b" seconds."""
    import time

    time.sleep(b)
    return a + b


@job
def create_detour(detour_job: Job):
    """Create a detour based on the passed Job."""
    from jobflow import Flow

    return Response(detour=Flow(detour_job))


@job
def self_replace(n: int):
    """Create a replace Job with the same job n times."""
    if n > 0:
        return Response(replace=self_replace(n - 1))

    return n


@job
def ignore_input(a: int) -> int:
    """
    Can receive an input, but ignores it.

    Allows to test flows with failed parents
    """
    return 1


@job
def current_jobdoc():
    from jobflow_remote.jobs.run import CURRENT_JOBDOC

    return CURRENT_JOBDOC.job_doc



from jobflow_remote.remote.host.base import BaseHost
from jobflow_remote.remote.host.local import LocalHost
from jobflow_remote.remote.host.remote import RemoteHost

__all__ = ("BaseHost", "LocalHost", "RemoteHost")



# ruff: noqa: PLW0602, PLW0603

from __future__ import annotations

import contextlib
import functools
import json
import logging
import uuid
from contextlib import contextmanager
from datetime import datetime, timedelta
from enum import Enum
from typing import TYPE_CHECKING, Callable, NoReturn

import typer
from click import ClickException
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Confirm
from rich.text import Text, TextType
from rich.tree import Tree
from typer.core import TyperCommand, TyperGroup

from jobflow_remote import ConfigManager, JobController
from jobflow_remote.config.base import ProjectUndefinedError
from jobflow_remote.jobs.daemon import DaemonError, DaemonManager, DaemonStatus

if TYPE_CHECKING:
    from cProfile import Profile

    from jobflow_remote.jobs.state import JobState

logger = logging.getLogger(__name__)


err_console = Console(stderr=True)
out_console = Console()


fmt_datetime = "%Y-%m-%d %H:%M"


# shared instances of the config manager and job controller, to avoid parsing
# the files multiple times. Needs to be initialized with the
# initialize_config_manager function.
_shared_config_manager: ConfigManager | None = None
_shared_job_controller: JobController | None = None

_profiler: Profile | None = None


def initialize_config_manager(*args, **kwargs) -> None:
    global _shared_config_manager
    _shared_config_manager = ConfigManager(*args, **kwargs)


def get_config_manager() -> ConfigManager:
    global _shared_config_manager
    if not _shared_config_manager:
        raise RuntimeError("The shared config manager needs to be initialized")
    return _shared_config_manager


def get_job_controller():
    global _shared_job_controller
    if _shared_job_controller is None:
        cm = get_config_manager()
        jc = JobController.from_project(cm.get_project())
        _shared_job_controller = jc

    return _shared_job_controller


def cleanup_job_controller() -> None:
    global _shared_job_controller
    if _shared_job_controller is not None:
        _shared_job_controller.close()
        # set to None again, in case it needs to be used again in the same
        # execution (e.g., in tests)
        _shared_job_controller = None


def start_profiling() -> None:
    global _profiler
    from cProfile import Profile

    _profiler = Profile()
    _profiler.enable()


def complete_profiling() -> None:
    global _profiler

    _profiler.disable()
    import pstats

    stats = pstats.Stats(_profiler).sort_stats("cumtime")
    stats.print_stats()


class SortOption(str, Enum):
    CREATED_ON = "created_on"
    UPDATED_ON = "updated_on"
    DB_ID = "db_id"


class SerializeFileFormat(str, Enum):
    JSON = "json"
    YAML = "yaml"
    TOML = "toml"


class IndexDirection(str, Enum):
    ASC = "asc"
    DESC = "desc"

    @property
    def as_pymongo(self):
        import pymongo

        return {
            IndexDirection.ASC: pymongo.ASCENDING,
            IndexDirection.DESC: pymongo.DESCENDING,
        }[self]


class ReportInterval(str, Enum):
    HOURS = "hours"
    DAYS = "days"
    WEEKS = "weeks"
    MONTHS = "months"
    YEARS = "years"


class ReprStr(str):
    r"""
    Helper class that overrides the standard __repr__ to return the string itself
    and not its repr().
    Used mainly to allow printing of strings with newlines instead of '\n' when
    repr is used in rich.
    """

    __slots__ = ()

    def __repr__(self) -> str:
        return self


def exit_with_error_msg(message: TextType, code: int = 1, **kwargs) -> NoReturn:
    kwargs.setdefault("style", "red")
    err_console.print(message, **kwargs)
    raise typer.Exit(code)


def exit_with_warning_msg(message: TextType, code: int = 0, **kwargs) -> NoReturn:
    kwargs.setdefault("style", "gold1")
    err_console.print(message, **kwargs)
    raise typer.Exit(code)


def print_success_msg(message: TextType = "operation completed", **kwargs) -> None:
    kwargs.setdefault("style", "green")
    out_console.print(message, **kwargs)


def check_incompatible_opt(d: dict) -> None:
    not_none = []
    for k, v in d.items():
        if v:
            not_none.append(k)

    if len(not_none) > 1:
        options_list = ", ".join(not_none)
        exit_with_error_msg(f"Options {options_list} are incompatible")


def check_query_incompatibility(query, incompatible_options):
    if query and any(opt is not None for opt in incompatible_options):
        exit_with_error_msg(
            "The --query option is incompatible with all the other filtering options"
        )


def check_at_least_one_opt(d: dict) -> None:
    not_none = []
    for k, v in d.items():
        if v:
            not_none.append(k)

    if len(not_none) > 1:
        options_list = ", ".join(d)
        exit_with_error_msg(
            f"At least one of the options {options_list} should be defined"
        )


def check_only_one_opt(d: dict) -> None:
    not_none = []
    for k, v in d.items():
        if v:
            not_none.append(k)

    if len(not_none) != 1:
        options_list = ", ".join(d)
        exit_with_error_msg(
            f"One and only one of the options {options_list} should be defined"
        )


@contextmanager
def loading_spinner(processing: bool = True):
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        if processing:
            progress.add_task(description="Processing...", total=None)
        yield progress


@contextmanager
def hide_progress(progress: Progress):
    """
    Hide the progress bar or spinning icon if an input is required from the user.

    Adapted from a related github issue for rich:
    https://github.com/Textualize/rich/issues/1535#issuecomment-1745297594

    Parameters
    ----------
    progress
        The Progress object in use
    """
    transient = progress.live.transient  # save the old value
    progress.live.transient = True
    progress.stop()
    progress.live.transient = transient  # restore the old value
    try:
        yield
    finally:
        # make space for the progress to use so it doesn't overwrite any previous lines
        progress.start()


def get_job_db_ids(job_db_id: str, job_index: int | None):
    if check_valid_uuid(job_db_id, raise_on_error=False):
        db_id = None
        job_id = job_db_id
    else:
        db_id = job_db_id
        job_id = None

    if job_index and db_id is not None:
        out_console.print(
            "The index is defined while a db_id is passed as an ID. Will be ignored",
            style="yellow",
        )

    return db_id, job_id


def get_job_ids_indexes(job_ids: list[str] | None) -> list[tuple[str, int]] | None:
    if not job_ids:
        return None
    job_ids_indexes = []
    for j in job_ids:
        split = j.split(":")
        if len(split) != 2 or not split[1].isnumeric():
            raise typer.BadParameter(
                "The job id should be in the format UUID:INDEX "
                "(e.g. e1d66c4f-81db-4fff-bda2-2bf1d79d5961:2). "
                f"Wrong format for {j}"
            )
        check_valid_uuid(split[0])
        job_ids_indexes.append((split[0], int(split[1])))

    return job_ids_indexes


def cli_error_handler(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (typer.Exit, typer.Abort, ClickException):
            raise  # Do not capture click or typer exceptions
        except ProjectUndefinedError:
            exit_with_error_msg(
                "The active project could not be determined and it is required to execute this command"
            )
        except Exception as e:
            from jobflow_remote import SETTINGS

            if SETTINGS.cli_full_exc:
                raise  # Reraise exceptions to print the full stacktrace
            exit_with_error_msg(
                f"An Error occurred during the command execution: {type(e).__name__} {getattr(e, 'message', str(e))}"
            )

    return wrapper


def check_valid_uuid(uuid_str, raise_on_error: bool = True) -> bool:
    try:
        uuid_obj = uuid.UUID(uuid_str)
        if str(uuid_obj) == uuid_str:
            return True
    except ValueError:
        pass

    if raise_on_error:
        raise typer.BadParameter(f"UUID {uuid_str} is in the wrong format.")
    return False


def str_to_dict(string: str | None) -> dict | None:
    if not string:
        return None

    try:
        dictionary = json.loads(string)
    except json.JSONDecodeError as exc:
        dictionary = {}
        for chunk in string.split(","):
            split = chunk.split("=")
            if len(split) != 2:
                raise typer.BadParameter(
                    f"Wrong format for dictionary-like field {string}"
                ) from exc

            dictionary[split[0]] = split[1]

    return dictionary


def get_start_date(start_date: datetime | None, days: int | None, hours: int | None):
    if start_date and (start_date.year, start_date.month, start_date.day) == (
        1900,
        1,
        1,
    ):
        now = datetime.now()
        start_date = start_date.replace(year=now.year, month=now.month, day=now.day)
        if start_date > now:
            start_date = start_date - timedelta(days=1)
    elif days:
        start_date = datetime.now() - timedelta(days=days)
    elif hours:
        start_date = datetime.now() - timedelta(hours=hours)

    return start_date


def execute_multi_jobs_cmd(
    single_cmd: Callable,
    multi_cmd: Callable,
    job_db_id: str | None = None,
    job_index: int | None = None,
    job_ids: list[str] | None = None,
    db_ids: str | list[str] | None = None,
    flow_ids: str | list[str] | None = None,
    states: JobState | list[JobState] | None = None,
    start_date: datetime | None = None,
    end_date: datetime | None = None,
    name: str | None = None,
    metadata: dict | None = None,
    days: int | None = None,
    hours: int | None = None,
    workers: list[str] | None = None,
    custom_query: dict | None = None,
    verbosity: int = 0,
    raise_on_error: bool = False,
    interactive: bool = True,
    **kwargs,
) -> None:
    """
    Utility function to execute a command on a single job or on a set of jobs.
    Checks the query options, determine whether the command is single or multi
    and call the corresponding function.

    Parameters
    ----------
    single_cmd
        A function that takes the job_id and job_index as arguments and performs an
        action on a single job.
    multi_cmd
        A function that takes a set of standard arguments and performs an
        action on multiple jobs.
    job_db_id
        The job id or db_id of a job to be considered. If specified, all the other
        query options should be disabled.
    job_index
        The index of the job in the DB.
    job_ids
        A list of job ids to be considered.
    db_ids
        A list of db ids to be considered.
    flow_ids
        A list of flow ids to be considered.
    states
        The states of the jobs to be considered.
    start_date
        The start date of the jobs to be considered.
    end_date
        The end date of the jobs to be considered.
    name
        The name of the jobs to be considered.
    metadata
        The metadata of the jobs to be considered.
    days
        Set the start_date based on the number of past days.
    hours
        Set the start_date based on the number of past hours.
    workers
        Workers associated with the jobs to be considered.
    custom_query
        A custom query to be used to filter the jobs.
    verbosity
        The verbosity of the output.
    raise_on_error
        If True, an error will be raised if the operation fails.
    interactive
        If True, a spinner will be shown even if the operation is not interactive while
        the operation is being performed.
    **kwargs : dict
        Additional arguments to be passed to the single_cmd and multi_cmd functions.
    """
    query_values = [
        job_ids,
        db_ids,
        flow_ids,
        states,
        start_date,
        end_date,
        name,
        metadata,
        days,
        hours,
        workers,
        custom_query,
    ]
    # if potentially interactive do not start the spinner.
    cm = get_config_manager()
    spinner_cm: contextlib.AbstractContextManager
    if interactive and cm.get_project().has_interactive_workers:
        spinner_cm = contextlib.nullcontext()
        out_console.print("Processing...")
    else:
        spinner_cm = loading_spinner()
    try:
        if job_db_id is not None:
            if any(query_values):
                msg = "If job_db_id is defined all the other query options should be disabled"
                exit_with_error_msg(msg)
            db_id, job_id = get_job_db_ids(job_db_id, job_index)
            with spinner_cm:
                modified_ids = single_cmd(
                    job_id=job_id, job_index=job_index, db_id=db_id, **kwargs
                )
                if not isinstance(modified_ids, (list, tuple)):
                    modified_ids = [] if modified_ids is None else [modified_ids]
                if not modified_ids:
                    exit_with_error_msg("Could not perform the requested operation")
        else:
            check_incompatible_opt(
                {"start_date": start_date, "days": days, "hours": hours}
            )
            check_incompatible_opt({"end_date": end_date, "days": days, "hours": hours})
            check_query_incompatibility(
                custom_query,
                [
                    job_ids,
                    db_ids,
                    flow_ids,
                    states,
                    start_date,
                    end_date,
                    name,
                    metadata,
                    days,
                    hours,
                    workers,
                ],
            )

            job_ids_indexes = get_job_ids_indexes(job_ids)
            start_date = get_start_date(start_date, days, hours)

            if not any(
                (
                    job_ids_indexes,
                    db_ids,
                    flow_ids,
                    states,
                    start_date,
                    end_date,
                    name,
                    metadata,
                    workers,
                    custom_query,
                )
            ):
                text = Text.from_markup(
                    "[yellow]No filter has been set. This will apply the change to all "
                    "the jobs in the DB. Proceed anyway?[/yellow]"
                )

                confirmed = Confirm.ask(text, default=False)
                if not confirmed:
                    raise typer.Exit(0)  # noqa: TRY301

            with spinner_cm:
                modified_ids = multi_cmd(
                    job_ids=job_ids_indexes,
                    db_ids=db_ids,
                    flow_ids=flow_ids,
                    states=states,
                    start_date=start_date,
                    end_date=end_date,
                    name=name,
                    metadata=metadata,
                    workers=workers,
                    custom_query=custom_query,
                    raise_on_error=raise_on_error,
                    **kwargs,
                )

        if verbosity:
            print_success_msg(f"Operation completed. Modified jobs: {modified_ids}")
        else:
            print_success_msg(f"Operation completed: {len(modified_ids)} jobs modified")
    except Exception:
        logger.exception("Error executing the operation")


def check_stopped_runner(error: bool = True) -> None:
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    try:
        with loading_spinner(processing=False) as progress:
            progress.add_task(description="Checking the Daemon status...", total=None)
            current_status = dm.check_status()

    except DaemonError as e:
        exit_with_error_msg(
            f"Error while checking the status of the daemon: {getattr(e, 'message', str(e))}"
        )

    # the current status is only local. Also check that the DB does not contain
    # a running_runner document that does not match the current machine.
    runner_stopped = False
    if current_status in (DaemonStatus.STOPPED, DaemonStatus.SHUT_DOWN):
        matching_error = dm.check_matching_runner(allow_missing=True)
        if matching_error:
            logger.warning(matching_error)
        else:
            runner_stopped = True
    if not runner_stopped:
        if error:
            exit_with_error_msg(
                f"The status of the daemon is {current_status.value}. "
                "The daemon should not be running while performing this operation"
            )
        else:
            text = Text.from_markup(
                "[red]The Runner is active. This operation may lead to "
                "inconsistencies in this case. Proceed anyway?[/red]"
            )

            confirmed = Confirm.ask(text, default=False)
            if not confirmed:
                raise typer.Exit(0)


def get_command_tree(
    app: TyperGroup,
    tree: Tree,
    show_options: bool,
    show_docs: bool,
    show_hidden: bool,
    max_depth: int | None,
    current_depth: int = 0,
    max_doc_lines: int | None = None,
) -> Tree:
    """
    Recursively build a tree representation of the command structure.

    Parameters
    ----------
    app
        The Typer app or command group to process.
    tree
        The Rich Tree object to add nodes to.
    show_options
        Whether to display command options.
    show_docs
        Whether to display command and option documentation.
    show_hidden
        Whether to display hidden commands.
    max_depth
        Maximum depth of the tree to display.
    current_depth
        Current depth in the tree (used for recursion).
    max_doc_lines
        If show_docs is True, the maximum number of lines showed from the
        documentation of each command/option.

    Returns
    -------
    Tree
        The populated Rich Tree object.
    """
    if max_depth is not None and current_depth >= max_depth:
        return tree

    for command_name, command in sorted(app.commands.items()):
        if not show_hidden and getattr(command, "hidden", False):
            continue

        command_str = f"[bold green]{command_name}[/bold green]"
        if (
            show_docs
            and isinstance(command, (TyperCommand, TyperGroup))
            and command.help
        ):
            command_str += f": {command.help}"
            if max_doc_lines:
                lines = command_str.splitlines()
                command_str = "\n".join(lines[:max_doc_lines])
                if len(lines) > max_doc_lines:
                    command_str += " ..."
        branch = tree.add(command_str)

        if isinstance(command, TyperGroup):
            get_command_tree(
                command,
                branch,
                show_options,
                show_docs,
                show_hidden,
                max_depth,
                current_depth + 1,
                max_doc_lines,
            )
        elif show_options:
            for param in command.params:
                param_str = f"[blue]{param.name}[/blue]"
                if param.default is not None:
                    default_value = param.default if param.default != "" else '""'
                    param_str += f" (default: {default_value})"
                if param.required:
                    param_str += " [red](required)[/red]"
                if show_docs and param.help:
                    param_str += f": {param.help}"
                if max_doc_lines:
                    lines = param_str.splitlines()
                    param_str = "\n".join(lines[:max_doc_lines])
                    if len(lines) > max_doc_lines:
                        param_str += " ..."
                branch.add(param_str)
    return tree


def tree_callback(
    ctx: typer.Context,
    value: bool,
):
    if value:
        main_app = ctx.command
        tree_title = f"[bold red]{ctx.command.name}[/bold red]"

        tree = Tree(tree_title)
        command_tree = get_command_tree(
            main_app,
            tree,
            show_options=False,
            show_docs=True,
            show_hidden=False,
            max_depth=None,
            max_doc_lines=1,
        )

        out_console.print(command_tree)
        raise typer.Exit(0)


def confirm_project_name(style: str | None = "red", exit_on_error: bool = True) -> bool:
    """
    Ask the user to insert the name of the current project to confirm that a specific operation
    should be performed.

    Parameters
    ----------
    style
        The style to use for the message. If None, no style is used.
    exit_on_error
        If True, exit the program with an error message if the user does not insert the correct
        project name. Otherwise, return False.

    Returns
    -------
    bool
        True if the user inserted the correct project name, False otherwise.
    """
    cm = get_config_manager()
    project = cm.get_project()
    msg = f"Insert the name of the project ({project.name}) to confirm that you want to proceed "
    if style:
        msg = f"[{style}]{msg}[/]"
    user_project_name = out_console.input(msg)
    if user_project_name.strip() != project.name:
        if exit_on_error:
            exit_with_error_msg("The input does not match the current project name")
        return False
    return True
